import sys
import inspect

from functools import partial

from .request import Request
from .response import Response
from .exceptions import (
    HTTPStatus,
    HTTPSuccess,
    HTTPNotFound,
    HTTPBadRequest,
    HTTPInternalServerError,
)
from .response_formatters import ResponseFormattersMixin

ROUTE_VERB = 0
ROUTE_ARG = 1
http_success = HTTPSuccess().status


class RouterMixin(ResponseFormattersMixin):
    __request_factory__ = Request
    __response_factory__ = Response
    __route_argument_char__ = ":"

    def __init__(self):
        self.request = None
        self.response = None
        self.routes = {}

    def begin_request(self):
        """ Hook """
        pass

    def begin_response(self):
        """ Hook """
        pass

    def end_response(self):
        """ Hook """
        pass

    def on_error(self, exc: HTTPStatus):
        """ Hook """
        pass

    def __getattr__(self, key):
        formatter = super().__getattribute__(f"format_{key}")
        return partial(self.route, formatter=formatter)

    def route(self, path, formatter=None, **kwargs):
        def decorator(func):
            app = self

            # Set formatter
            func.__gongish_formatter__ = partial(
                formatter or self.__class__.__default_formatter__, **kwargs
            )

            # Get parameters
            func.__gongish_route_params__ = tuple(
                [
                    (param.name, param.annotation)
                    for param in inspect.signature(func).parameters.values()
                ]
            )

            # Distribute
            resources = path.split("/")
            parent = app.routes
            for resource in resources:

                if resource.startswith(app.__route_argument_char__):
                    resource = ROUTE_ARG

                if resource not in parent.keys():
                    parent[resource] = {}

                parent = parent[resource]

            if ROUTE_VERB not in parent.keys():
                parent[ROUTE_VERB] = {func.__name__: func}
            else:
                parent[ROUTE_VERB][func.__name__] = func

        return decorator

    def handle_exception(self, exc, start_response):
        if isinstance(exc, HTTPStatus):
            exc_ = exc
            exc_info = None

        else:
            exc_ = HTTPInternalServerError()
            exc_info = sys.exc_info()
            self.__logger__.exception(exc_.text, exc_info=True)

        if exc_.__keep_headers__:
            self.response.headers.update(exc_.headers)
        else:
            self.response.headers = exc_.headers

        self.response.charset = "utf-8"
        self.response.type = "plain/text"
        self.response.body = exc_.render(debug=self.config.debug)

        self.on_error(exc_)  # hook

        body = self.response.conclude()

        start_response(
            exc_.status, list(self.response.headers.items()), exc_info
        )

        self.end_response()  # hook

        return body

    def dispatch(self, path: str, verb: str):
        """
        Dispatch path and return handler function
        """

        # Split path
        resources = path.split("/")
        route = self.routes
        route_keys = route.keys()
        route_args = []
        steps = 0
        for resource in resources:
            if resource in route_keys:
                route = route[resource]
                route_keys = route.keys()
                steps += 1

            elif ROUTE_ARG in route_keys:
                # Extract arguments
                route = route[ROUTE_ARG]
                route_keys = route.keys()
                route_args.append(resource)
                steps += 1

        if steps != len(resources) or verb not in route[ROUTE_VERB].keys():
            raise HTTPNotFound

        handler = route[ROUTE_VERB][verb]

        # Validate parameters
        for idx, param in enumerate(handler.__gongish_route_params__):
            name, annotation = param

            if annotation == inspect._empty:
                continue

            # Call the annotation
            try:
                route_args[idx] = annotation(route_args[idx])

            except Exception:
                raise HTTPBadRequest(f"Invalid Parameter `{name}`")

        return handler, route_args

    def __call__(self, environ, start_response):
        """ Application WSGI entry """
        self.request = request = self.__request_factory__(environ)
        self.response = response = self.__response_factory__()
        self.begin_request()  # hook

        try:
            # Call handler
            handler, route_args = self.dispatch(request.path, request.verb)
            response.body = handler(*route_args)
            response.status = http_success

            # Format response
            handler.__gongish_formatter__(request, response)

        except Exception as exc:
            return self.handle_exception(exc, start_response)

        self.begin_response()  # hook

        # Cookies
        cookie = request.cookies.output()
        if cookie:
            for line in cookie.split("\r\n"):
                response.headers.add(*line.split(": ", 1))

        body = response.conclude()

        start_response(response.status, list(response.headers.items()))

        self.end_response()  # hook
        return body



from collections import OrderedDict


class HeaderSet(OrderedDict):  # pragma: nocover
    def __init__(self, items=None):
        super().__init__()
        for i in items or []:
            self.add(i)

    def add(self, k, *args):
        if isinstance(k, str):
            values = []
            if ":" in k:
                k, v = k.split(":", 1)
                values += [i.strip() for i in v.split(";")]

        else:
            k, v = k
            values = [v]

        if args:
            values += args

        self[k] = "; ".join(values)

    def __getitem__(self, k):
        return super().__getitem__(k.lower())

    def __setitem__(self, k, v):
        return super().__setitem__(k.lower(), v)

    def __delitem__(self, k):
        return super().__delitem__(k)

    def __iter__(self):
        for k, v in self.items():
            yield f"{k}: {v}"

    def __str__(self):
        return "\r\n".join(self)

    def __iadd__(self, other):
        for i in other:
            self.add(i)

        return self

    def load_from_wsgi_environ(self, environ):
        for key, value in environ.items():
            if key.startswith("HTTP_") and key not in (
                "HTTP_CONTENT_TYPE",
                "HTTP_CONTENT_LENGTH",
            ):
                header_name = key[5:].replace("_", "-")
                self[header_name] = value

            elif key in ("CONTENT_TYPE", "CONTENT_LENGTH") and value:
                header_name = key.replace("_", "-")
                self[header_name] = value

        return self


class LazyAttribute:  # pragma: nocover
    """
    ``LazyAttribute`` decorator is intended to promote a
    function call to object attribute. This means the
    function is called once and replaced with
    returned value.
    >>> class A:
    ...     def __init__(self):
    ...         self.counter = 0
    ...     @LazyAttribute
    ...     def count(self):
    ...         self.counter += 1
    ...         return self.counter
    >>> a = A()
    >>> a.count
    1
    >>> a.count
    1
    """

    __slots__ = ("f",)

    def __init__(self, f):
        self.f = f

    def __get__(self, obj, t=None):
        f = self.f
        if obj is None:
            return f
        val = f(obj)
        setattr(obj, f.__name__, val)
        return val



from .helpers import HeaderSet


class Response:
    def __init__(self):
        self.headers = HeaderSet()
        self.status = "200 OK"
        self.length = None
        self.body = None
        self._type = None
        self._charset = None

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, val):
        self._type = val
        self.headers["content-type"] = self.contenttype

    @property
    def charset(self):
        return self._charset

    @charset.setter
    def charset(self, val):
        self._charset = val
        self.headers["content-type"] = self.contenttype

    @property
    def contenttype(self):
        if not self._type:
            return None

        result = self._type
        if self._charset:
            result += f"; charset={self._charset}"

        return result

    def conclude(self):
        body = self.body
        if body is None:
            body = []

        elif isinstance(body, (str, bytes)):
            body = [body]

        if self.charset:
            body = [i.encode(self.charset) for i in body]

        if self.length is None:
            self.length = sum(len(i) for i in body)

        return body



try:
    import ujson as jsonlib
except ImportError:  # pragma: no cover
    import json as jsonlib


class ResponseFormattersMixin:
    @staticmethod
    def format_text(request, response):
        # Format in plain text
        response.type = "text/plain"
        response.charset = "utf-8"
        response.body = str(response.body)

    @staticmethod
    def format_json(request, response, indent=None):
        # Format in JSON
        response.type = "application/json"
        response.charset = "utf-8"
        response.body = jsonlib.dumps(response.body, indent=indent)

    @staticmethod
    def format_binary(request, response):
        # Format in Binary
        response.type = "application/octet-stream"

    __default_formatter__ = format_text
